(* (C)        Cuihtlauac Alvarado, 2000            *)
(*            France Telecom R & D                 *)
(* $Id: otags.ml,v 1.21 2002/11/13 16:00:49 alvaracu Exp $ *)

let otags = "otags"

let release = "3.06.5"

let print_version _ = print_endline (otags ^ " " ^ release)

let default_parsers = [
  "pa_o.cmo";
  "pa_op.cmo";
]

type editor = Emacs | Vi

let pr_tags = function
  | Emacs -> "pr_emacs_tags.cma"
  | Vi -> "pr_vi_tags.cma"

module Suffix = struct
  
  let list = ref [ ".mli"; ".ml" ]

  let contents () = !list
  let clear _ = list := []
  let add s = list := s :: !list
  let remove s =
    let rec loop = function
      | [] -> []
      | h :: t -> if h = s then t else h :: loop t
    in list := loop !list

end

(* Parsing arguments *)
let verbose, output, recursive, editor, targets, suffixes, camlp4, parsers = 
  let rverbose = ref false in
  let routput = ref "" in
  let rrecursive = ref false in
  let reditor = ref Emacs in
  let rcamlp4 = ref "camlp4" in

  let q = Mklist.create () in
  let add_target s = Mklist.add s q in
  let targets () = Mklist.from q in

  let p = Mklist.create () in
  let add_parser s = Mklist.add s p in
  let parsers () = Mklist.from p in

  let usage = [
    ("-v", Arg.Set rverbose, "            verbose, display debug messages");
    ("-o", Arg.String ((:=) routput), "<str>       output tags file (default: TAGS) (use - for standard output)");
    ("-sc", Arg.Unit Suffix.clear, "           suffix list: clear (default: [.mli; .ml])");
    ("-sa", Arg.String Suffix.add, "<str>      suffix list: add target");
    ("-sr", Arg.String Suffix.remove, "<str>      suffix list: remove target");
    ("-r", Arg.Set rrecursive, "            enable recursive directory search");
    ("-f", Arg.Clear rrecursive, "            disable recursive directory search (default)");
    ("-vi", Arg.Unit (fun _ -> reditor := Vi), "           generate vi tags");
    ("-emacs", Arg.Unit (fun _ -> reditor := Emacs), "        generate emacs tags (default)");
    ("-camlp4", Arg.String ((:=) rcamlp4), "<str>  camlp4 command (default: camlp4)");
    ("-pa", Arg.String add_parser, "<str>      add camlp4 parser (default: pa_o.cmo; pa_op.cmo)");
    ("-version", Arg.Unit print_version, "      display otags version number");
  ] in
  Arg.parse usage add_target
    "Available commands:\n  -help          display this help message";
  let parsers = let p = parsers() in
    match p with 
    | [] -> default_parsers
    | _ :: _ -> p in
  let output =  
     if !routput = "" then 
        match !reditor with
        | Emacs -> "TAGS"
        | Vi -> "tags"
     else
        !routput in
  !rverbose, output, !rrecursive, !reditor, 
  targets (), Suffix.contents (), !rcamlp4, parsers

let camlp4_libs pr_tags = parsers @ [ pr_tags ]

let debug prog args ofile =
  if verbose then begin
    List.iter (Printf.printf "%s ") (prog::args);
    begin 
      match ofile with
      | None -> ()
      | Some filename -> Printf.printf "> %s " filename 
    end;
    print_newline ();
    flush stdout
  end

let error s =
  begin
    prerr_string s;
    flush stderr
  end

let rec locate file_name = function
  | [] -> file_name
  | prefix :: pos_list ->
      let full_name = prefix ^ "/" ^ file_name in
      try (Unix.access full_name [ Unix.F_OK; Unix.R_OK ]; full_name)
      with Unix.Unix_error _ -> locate file_name pos_list
  
let pr_tags_loc = locate (pr_tags editor) [ "."; Config.lib ]

let system prog args ofile =
  debug prog args ofile;
  let fd =
    match ofile with
    | None | Some "-" -> Unix.stdout
    | Some filename -> Unix.descr_of_out_channel (open_out filename) in
  let pid = Unix.create_process prog (Array.of_list (prog::args)) Unix.stdin fd Unix.stderr in
  let _ = Unix.wait () in
  ()
        
module Files : 
  sig 
    val ls : string -> string list
    val sort : string list -> string list
  end =
struct
  let ls dir =
    let l = Mklist.create () in
    let dir_hdl = Unix.opendir dir in
    begin
      ignore (Unix.readdir dir_hdl); (* Skip . *)
      ignore (Unix.readdir dir_hdl); (* Skip .. *)
      try while true do (* Mklist.add (Unix.readdir dir_hdl) l done *)
        let s = (Unix.readdir dir_hdl) in
        if s.[0] != '.' then Mklist.add s l
      done
      with End_of_file -> Unix.closedir dir_hdl
    end;
    Mklist.from l
      
  let has_suff s f = Filename.check_suffix f s
  let chop f = try Filename.chop_extension f with Invalid_argument _ -> f
  let rec first_suff f1 f2 = function
    | [] -> 0
    | s :: ls -> 
	if has_suff s f1 then -1 
	else if has_suff s f2 then 1
	else first_suff f1 f2 ls
  let cmp f1 f2 = 
    let c1 = chop f1 in
    let c2 = chop f2 in
    let c = compare c1 c2 in
    if c = 0 then 
      if f1 = f2 then 0
      else first_suff f1 f2 suffixes
    else c

  let sort = List.sort cmp
end

(* tfl = tag file list *)
let process_file tfl file =
  if (List.exists) (Filename.check_suffix file) suffixes then
    let tagfile = Filename.temp_file "otags" "" in
    let _ = Unix.putenv Argcamlp4.tmp tagfile in
    let _ = system camlp4 (camlp4_libs pr_tags_loc @ [file]) None in
    tagfile :: tfl
  else tfl

let rec process_dir tfl dir =
  let files = Files.sort (Files.ls dir) in
  let rec aux tfl = function
    | [] -> tfl
    | f :: lf -> 
	let tfl1 = process_arg recursive tfl (Filename.concat dir f) in
	aux tfl1 lf 
  in aux tfl files
and process_arg thisrecur tfl arg =
  match (Unix.stat arg).Unix.st_kind with
  | Unix.S_DIR ->
      if thisrecur then process_dir tfl arg
      else tfl
  | _ -> process_file tfl arg

let rec process_arg_list tfl = function
  | [] -> tfl
  | arg :: la ->
      let tfl1 = process_arg true tfl arg in
      process_arg_list tfl1 la

let main _ = 
  let tfl = process_arg_list [] targets in
  system "cat" ("/dev/null" :: List.rev tfl) (Some output);
  system "rm" ("-f" :: tfl) None

let _ = try main () with 
  | Unix.Unix_error (code, funct, arg) ->
      begin
        error (otags ^ ": Unix error\n");
        error ("  Syscall: " ^ funct ^ "\n");
        error ("  Arg:     " ^ arg ^ "\n");
        error ("  Error:   " ^ Unix.error_message code ^ "\n");
      end
  | _ as e -> 
      begin
        error (otags ^ ": Fatal error\n");
        raise e;
      end 
